<?php

namespace Okawari\JobLogging;

use App\Contracts\NamedJob;
use App\Contracts\OwnedJob;
use App\Models\JobLog;
use App\Services\JobLogger;
use Illuminate\Queue\Events\JobFailed;
use Illuminate\Queue\Events\JobProcessed;
use Illuminate\Queue\Events\JobProcessing;
use Illuminate\Queue\Jobs\SyncJob;
use Illuminate\Support\Facades\Event;
use Illuminate\Support\Facades\Log;
use Illuminate\Support\ServiceProvider;

class JobLogProvider extends ServiceProvider
{
    /**
     * Register services.
     */
    public function register(): void
    {

        $driver = Log::getDefaultDriver();

        Event::listen(function (JobProcessing $event) {
            /** @var \Monolog\Logger $logger */
            $logger = Log::getLogger();

            $job = $event->job;
            $owner = $job instanceof OwnedJob ? $job->getOwner() : null;
            $name = $job instanceof NamedJob ? $job->getJobLogName() : $event->job->getName();

            if ($job instanceof SyncJob) {
                $job = unserialize(data_get(json_decode($job->getRawBody()), 'data.command'));
                $owner = $job instanceof OwnedJob ? $job->getOwner() : $owner;
                $name = $job instanceof NamedJob ? $job->getJobLogName() : $name;
            }

            $jobLog = app(JobLog::class)::create([
                'job_id' => $event->job->getJobId() ?: '',
                'class' => get_class($event->job),
                'name' => $name,
                'owner_type' => $owner ? get_class($owner) : null,
                'owner_id' => $owner?->id,
                'output' => '',
                'started_at' => now(),
            ]);

            $logger->pushProcessor(new JobLogger($jobLog));
        });

        Event::listen(function (JobFailed $event) {
            /** @var \Monolog\Logger $logger */
            $jobLogger = $this->getJobLogger();

            if (!$jobLogger) {
                return;
            }

            $job = $event->job;
            $logLine = $jobLogger->log;
            $owner = $job instanceof OwnedJob ? $job->getOwner() : null;
            $name = $job instanceof NamedJob ? $job->getJobLogName() : $event->job->getName();
            if ($owner) $logLine->owner()->associate($owner);
            if ($name) $logLine->name = $name ?: $logLine->name;

            $logLine->output .= "[EXCEPTION] " . now()->toIso8601ZuluString() . ": {$event->exception->getMessage()}";
            $success = $event->job?->hasFailed() === false;
            $logLine->success = $success;
            $logLine->ended_at = now();
            $logLine->save();
        });

        Event::listen(function (JobProcessed|JobFailed $event) {
            try {
                /** @var \Monolog\Logger $logger */
                $jobLogger = $this->getJobLogger();
                if (!$jobLogger) {
                    return;
                }

                $job = $event->job;
                $logLine = $jobLogger->log;

                $owner = $job instanceof OwnedJob ? $job->getOwner() : null;
                $name = $job instanceof NamedJob ? $job->getJobLogName() : $event->job->getName();
                if ($owner) $logLine->owner()->associate($owner);
                if ($name) $logLine->name = $name ?: $logLine->name;

                $logLine->success = !$event->job->hasFailed();
                $logLine->ended_at = now();
                $logLine->save();
            } catch (\Throwable $t) {
                dd($t, debug_backtrace());
            }
        });
    }

    /**
     * Bootstrap services.
     */
    public function boot(): void
    {

    }

    /**
     * @return array
     */
    function getJobLogger(): ?JobLogger
    {
        $logger = Log::driver()->getLogger();
        $jobLogger = $logger->popProcessor();
        if ($jobLogger instanceof JobLogger) {
            return $jobLogger;
        } else {
            $logger->pushProcessor($jobLogger);
            return null;
        }
    }
}
